@@ -6,6 +6,7 @@
 # Copyright (c) 2008-2009 Yu Yuwei <acevery@gmail.com>
 # Copyright (c) 2009-2014 Caius "kaio" CHANCE <me@kaio.net>
 # Copyright (c) 2012-2018 Mike FABIAN <mfabian@redhat.com>
+# Copyright (c) 2019      Peng Wu <alexepico@gmail.com>
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -32,11 +33,14 @@ import sqlite3
 import uuid
 import time
 import re
+import logging
 import chinese_variants
 if sys.version_info < (3, 0, 0):
     reload(sys)
     sys.setdefaultencoding('utf-8')
 
+LOGGER = logging.getLogger('ibus-table')
+
 DEBUG_LEVEL = int(0)
 
 DATABASE_VERSION = '1.00'
@@ -85,14 +89,13 @@ class ImeProperties:
         if default_properties is None:
             default_properties = {}
         if not db:
-            return None
+            return
         self.ime_property_cache = default_properties
         sqlstr = 'SELECT attr, val FROM main.ime;'
         try:
             results = db.execute(sqlstr).fetchall()
         except:
-            import traceback
-            traceback.print_exc()
+            LOGGER.exception('Cannot get ime properties from database')
         for result in results:
             self.ime_property_cache[result[0]] = result[1]
 
@@ -164,9 +167,7 @@ class TabSqliteDb:
             self.db.execute('PRAGMA journal_size_limit = 1000000;')
             self.db.execute('PRAGMA synchronous = NORMAL;')
         except:
-            import traceback
-            traceback.print_exc()
-            print('Error while initializing database.')
+            LOGGER.exception('Error while initializing database')
         # create IME property table
         self.db.executescript(
             'CREATE TABLE IF NOT EXISTS main.ime (attr TEXT, val TEXT);')
@@ -198,6 +199,7 @@ class TabSqliteDb:
             'def_full_width_letter':'false',
             'user_can_define_phrase':'false',
             'pinyin_mode':'false',
+            'suggestion_mode':'false',
             'dynamic_adjust':'false',
             'auto_select':'false',
             'auto_commit':'false',
@@ -255,6 +257,16 @@ class TabSqliteDb:
                 + 'is it an outdated database?')
             self.dynamic_adjust = False
 
+        self.suggestion_mode = self.ime_properties.get('suggestion_mode')
+        if self.suggestion_mode:
+            self.suggestion_mode = bool(
+                self.suggestion_mode.lower() == u'true')
+        else:
+            print(
+                'Could not find "suggestion_mode" entry from database, '
+                + 'is it an outdated database?')
+            self.suggestion_mode = False
+
         self.rules = self.get_rules()
         self.possible_tabkeys_lengths = self.get_possible_tabkeys_lengths()
         self.startchars = self.get_start_chars()
@@ -301,9 +313,8 @@ class TabSqliteDb:
                     os.makedirs(tables_path, exist_ok=True)
             user_db = path.join(tables_path, user_db)
             if not path.exists(user_db):
-                sys.stderr.write(
-                    'The user database %(udb)s does not exist yet.\n'
-                    % {'udb': user_db})
+                LOGGER.debug(
+                    'The user database %s does not exist yet.', user_db)
             else:
                 try:
                     desc = self.get_database_desc(user_db)
@@ -314,71 +325,69 @@ class TabSqliteDb:
                             or (self.get_number_of_columns_of_phrase_table(
                                 user_db)
                                 != len(phrase_table_column_names))):
-                        sys.stderr.write(
-                            'The user database %s seems to be incompatible.\n'
-                            % user_db)
+                        LOGGER.debug(
+                            'The user database %s seems to be incompatible.',
+                            user_db)
                         if desc is None:
-                            sys.stderr.write(
+                            LOGGER.debug(
                                 'There is no version information in '
-                                + 'the database.\n')
+                                'the database.')
                             self.old_phrases = self.extract_user_phrases(
                                 user_db, old_database_version='0.0')
                         elif desc["version"] != DATABASE_VERSION:
-                            sys.stderr.write(
+                            LOGGER.debug(
                                 'The version of the database does not match '
-                                + '(too old or too new?).\n'
-                                'ibus-table wants version=%s\n'
-                                % DATABASE_VERSION
-                                + 'But the  database actually has version=%s\n'
-                                % desc['version'])
+                                '(too old or too new?). '
+                                'ibus-table wants version=%s '
+                                'But the  database actually has version=%s',
+                                DATABASE_VERSION, desc['version'])
                             self.old_phrases = self.extract_user_phrases(
                                 user_db, old_database_version=desc['version'])
                         elif (self.get_number_of_columns_of_phrase_table(
                                 user_db)
                               != len(phrase_table_column_names)):
-                            sys.stderr.write(
+                            LOGGER.debug(
                                 'The number of columns of the database '
-                                + 'does not match.\n'
-                                + 'ibus-table expects %s columns.\n'
-                                % len(phrase_table_column_names)
-                                + 'But the database actually has %s columns.\n'
-                                % self.get_number_of_columns_of_phrase_table(
-                                    user_db)
-                                + 'But the versions of the databases are '
-                                + 'identical.\n'
-                                + 'This should never happen!\n')
+                                'does not match. '
+                                'ibus-table expects %s columns. '
+                                'But the database actually has %s columns. '
+                                'But the versions of the databases are '
+                                'identical. '
+                                'This should never happen!',
+                                len(phrase_table_column_names),
+                                self.get_number_of_columns_of_phrase_table(
+                                    user_db))
                             self.old_phrases = None
                         from time import strftime
                         timestamp = strftime('-%Y-%m-%d_%H:%M:%S')
-                        sys.stderr.write(
-                            'Renaming the incompatible database to "%s".\n'
-                            % user_db+timestamp)
+                        LOGGER.debug(
+                            'Renaming the incompatible database to "%s".',
+                            user_db+timestamp)
                         if os.path.exists(user_db):
                             os.rename(user_db, user_db+timestamp)
                         if os.path.exists(user_db+'-shm'):
                             os.rename(user_db+'-shm', user_db+'-shm'+timestamp)
                         if os.path.exists(user_db+'-wal'):
                             os.rename(user_db+'-wal', user_db+'-wal'+timestamp)
-                        sys.stderr.write(
-                            'Creating a new, empty database "s".\n'
-                            % user_db)
+                        LOGGER.debug(
+                            'Creating a new, empty database "s".', user_db)
                         self.init_user_db(user_db)
-                        sys.stderr.write(
+                        LOGGER.debug(
                             'If user phrases were successfully recovered from '
-                            + 'the old,\n'
-                            + 'incompatible database, they will be used to '
-                            + 'initialize the new database.\n')
+                            'the old, '
+                            'incompatible database, they will be used to '
+                            'initialize the new database.')
                     else:
-                        sys.stderr.write(
-                            'Compatible database %s found.\n' % user_db)
+                        LOGGER.debug(
+                            'Compatible database %s found.', user_db)
                 except:
-                    import traceback
-                    traceback.print_exc()
+                    LOGGER.exception(
+                        'Unexpected error trying to find user database')
 
         # open user phrase database
         try:
-            sys.stderr.write(
-                'Connect to the database %(name)s.\n' %{'name': user_db})
+            LOGGER.debug(
+                'Connect to the database %s.', user_db)
             self.db.executescript('''
                 ATTACH DATABASE "%s" AS user_db;
                 PRAGMA user_db.encoding = "UTF-8";
@@ -391,19 +400,18 @@ class TabSqliteDb:
                 PRAGMA user_db.synchronous = NORMAL;
             ''' % user_db)
         except:
-            sys.stderr.write('Could not open the database %s.\n' % user_db)
+            LOGGER.debug('Could not open the database %s.', user_db)
             from time import strftime
             timestamp = strftime('-%Y-%m-%d_%H:%M:%S')
-            sys.stderr.write('Renaming the incompatible database to "%s".\n'
-                             % user_db+timestamp)
+            LOGGER.debug('Renaming the incompatible database to "%s".',
+                         user_db+timestamp)
             if os.path.exists(user_db):
                 os.rename(user_db, user_db+timestamp)
             if os.path.exists(user_db+'-shm'):
                 os.rename(user_db+'-shm', user_db+'-shm'+timestamp)
             if os.path.exists(user_db+'-wal'):
                 os.rename(user_db+'-wal', user_db+'-wal'+timestamp)
-            sys.stderr.write('Creating a new, empty database "%s".\n'
-                             % user_db)
+            LOGGER.debug('Creating a new, empty database "%s".', user_db)
             self.init_user_db(user_db)
             self.db.executescript('''
                 ATTACH DATABASE "%s" AS user_db;
@@ -432,8 +440,7 @@ class TabSqliteDb:
             try:
                 self.db.executemany(sqlstr, sqlargs)
             except:
-                import traceback
-                traceback.print_exc()
+                LOGGER.exception('Error inserting old phrases')
             self.db.commit()
             self.db.execute('PRAGMA wal_checkpoint;')
 
@@ -446,11 +453,9 @@ class TabSqliteDb:
             user_freq=0, database='user_db', commit=True):
         '''update phrase freqs'''
         if DEBUG_LEVEL > 1:
-            sys.stderr.write(
-                'update_phrase() tabkeys=%(t)s phrase=%(p)s '
-                % {'t': tabkeys, 'p': phrase}
-                + 'user_freq=%(u)s database=%(d)s\n'
-                % {'u': user_freq, 'd': database})
+            LOGGER.debug(
+                'tabkeys=%s phrase=%s user_freq=%s database=%s',
+                tabkeys, phrase, user_freq, database)
         if not tabkeys or not phrase:
             return
         sqlstr = '''
@@ -466,8 +471,7 @@ class TabSqliteDb:
                 self.db.commit()
             self.invalidate_phrases_cache(tabkeys)
         except:
-            import traceback
-            traceback.print_exc()
+            LOGGER.exception('Unexpected error updating phrase in user_db.')
 
     def sync_usrdb(self):
         '''
@@ -484,7 +488,7 @@ class TabSqliteDb:
         Make the phrases cache empty
         '''
         if DEBUG_LEVEL > 1:
-            sys.stderr.write("reset_phrases_cache()\n")
+            LOGGER.debug('reset_phrases_cache()')
         self._phrases_cache = {}
 
     def invalidate_phrases_cache(self, tabkeys=u''):
@@ -496,7 +500,7 @@ class TabSqliteDb:
         :type tabkeys: String
         '''
         if DEBUG_LEVEL > 1:
-            sys.stderr.write("invalidate_phrases_cache()\n")
+            LOGGER.debug('invalidate_phrases_cache()')
         for i in range(1, self._mlen + 1):
             if self._phrases_cache.get(tabkeys[0:i]):
                 self._phrases_cache.pop(tabkeys[0:i])
@@ -506,7 +510,7 @@ class TabSqliteDb:
         Load phrases cache from disk
         '''
         if DEBUG_LEVEL > 1:
-            sys.stderr.write('load_phrases_cache()\n')
+            LOGGER.debug('load_phrases_cache()')
         try:
             import json
             self._phrases_cache = json.load(open(self.cache_path, 'r'))
@@ -515,24 +519,21 @@ class TabSqliteDb:
                 self._phrases_cache = {}
         except FileNotFoundError:
             if DEBUG_LEVEL > 1:
-                sys.stderr.write(
-                    'load_phrases_cache() File %s not found\n'
-                    % self.cache_path)
+                LOGGER.debug(
+                    'File %s not found', self.cache_path)
         except PermissionError:
             if DEBUG_LEVEL > 1:
-                sys.stderr.write(
-                    'load_phrases_cache() Permission error reading %s\n'
-                    % self.cache_path)
+                LOGGER.debug(
+                    'Permission error reading %s', self.cache_path)
         except:
-            sys.stderr.write('load_phrases_cache() Unknown error reading %s\n'
-                             % self.cache_path)
+            LOGGER.debug('Unknown error reading %s', self.cache_path)
 
     def save_phrases_cache(self):
         '''
         Save phrases cache from disk
         '''
         if DEBUG_LEVEL > 1:
-            sys.stderr.write('save_phrases_cache()\n')
+            LOGGER.debug('save_phrases_cache()')
         try:
             import json
             self._phrases_cache['serial_number'] = self._snum
@@ -542,8 +543,7 @@ class TabSqliteDb:
             json.dump(self._phrases_cache, open(_cache_path, 'w'))
             os.replace(_cache_path, self.cache_path)
         except:
-            import traceback
-            traceback.print_exc()
+            LOGGER.exception('Unexpected error in save_phrases_cache().')
 
     def is_chinese(self):
         '''
@@ -627,6 +627,11 @@ class TabSqliteDb:
             (pinyin TEXT, zi TEXT, freq INTEGER);
             ''' % database
             self.db.execute(sqlstr)
+            sqlstr = '''
+            CREATE TABLE IF NOT EXISTS %s.suggestion
+            (phrase TEXT, freq INTEGER);
+            ''' %database
+            self.db.execute(sqlstr)
 
         sqlstr = '''
         CREATE TABLE IF NOT EXISTS %s.phrases
@@ -700,8 +705,7 @@ class TabSqliteDb:
                 else:
                     print('not a legal rule: "%s"' %rule)
         except Exception:
-            import traceback
-            traceback.print_exc()
+            LOGGER.exception('Unexpected error in get_rules().')
         return rules
 
     def get_possible_tabkeys_lengths(self):
@@ -731,15 +735,14 @@ class TabSqliteDb:
         if self.rules:
             max_len = self.rules["above"]
             return [len(self.rules[x]) for x in range(2, max_len+1)][:]
-        else:
-            try:
-                least_commit_len = int(
-                    self.ime_properties.get('least_commit_length'))
-            except (TypeError, ValueError):
-                least_commit_len = 0
-            if least_commit_len > 0:
-                return list(range(least_commit_len, self._mlen + 1))
-            return []
+        try:
+            least_commit_len = int(
+                self.ime_properties.get('least_commit_length'))
+        except (TypeError, ValueError):
+            least_commit_len = 0
+        if least_commit_len > 0:
+            return list(range(least_commit_len, self._mlen + 1))
+        return []
 
     def get_start_chars(self):
         '''return possible start chars of IME'''
@@ -770,8 +773,7 @@ class TabSqliteDb:
 
         '''
         if DEBUG_LEVEL > 1:
-            sys.stderr.write("add_phrases() len(phrases)=%s\n"
-                             %len(phrases))
+            LOGGER.debug('len(phrases)=%s', len(phrases))
         insert_sqlstr = '''
         INSERT INTO %(database)s.phrases
         (tabkeys, phrase, freq, user_freq)
@@ -796,11 +798,10 @@ class TabSqliteDb:
         (tabkeys, phrase, freq ,user_freq)
         '''
         if DEBUG_LEVEL > 1:
-            sys.stderr.write(
-                'add_phrase tabkeys=%(t)s phrase=%(p)s '
-                % {'t': tabkeys, 'p': phrase}
-                + 'freq=%(f)s user_freq=%(u)s\n'
-                % {'f': freq, 'u': user_freq})
+            LOGGER.debug(
+                'add_phrase tabkeys=%s phrase=%s '
+                'freq=%s user_freq=%s',
+                tabkeys, phrase, freq, user_freq)
         if not tabkeys or not phrase:
             return
         select_sqlstr = '''
@@ -813,12 +814,10 @@ class TabSqliteDb:
             # there is already such a phrase, i.e. add_phrase was called
             # in error, do nothing to avoid duplicate entries.
             if DEBUG_LEVEL > 1:
-                sys.stderr.write(
-                    'add_phrase() '
-                    + 'select_sqlstr=%(sql)s select_sqlargs=%(arg)s '
-                    % {'sql': select_sqlstr, 'arg': select_sqlargs}
-                    + 'already there!: results=%(r)s \n'
-                    % {'r': results})
+                LOGGER.debug(
+                    'select_sqlstr=%(sql)s select_sqlargs=%(arg)s '
+                    'already there!: results=%(r)s ',
+                    select_sqlstr, select_sqlargs, results)
             return
 
         insert_sqlstr = '''
@@ -832,17 +831,16 @@ class TabSqliteDb:
             'freq': freq,
             'user_freq': user_freq}
         if DEBUG_LEVEL > 1:
-            sys.stderr.write(
-                'add_phrase() insert_sqlstr=%(sql)s insert_sqlargs=%(arg)s\n'
-                % {'sql': insert_sqlstr, 'arg': insert_sqlargs})
+            LOGGER.debug(
+                'insert_sqlstr=%s insert_sqlargs=%s',
+                insert_sqlstr, insert_sqlargs)
         try:
             self.db.execute(insert_sqlstr, insert_sqlargs)
             if commit:
                 self.db.commit()
             self.invalidate_phrases_cache(tabkeys)
         except:
-            import traceback
-            traceback.print_exc()
+            LOGGER.exception('Unexpected error in add_phrase()')
 
     def add_goucima(self, goucimas):
         '''Add goucima into database, goucimas is iterable object
@@ -860,8 +858,7 @@ class TabSqliteDb:
             self.db.commit()
             self.db.execute('PRAGMA wal_checkpoint;')
         except:
-            import traceback
-            traceback.print_exc()
+            LOGGER.exception('Unexpected error in add_goucima().')
 
     def add_pinyin(self, pinyins, database='main'):
         '''Add pinyin to database, pinyins is a iterable object
@@ -883,12 +880,30 @@ class TabSqliteDb:
                 self.db.execute(
                     sqlstr, {'pinyin': pinyin, 'zi': zi, 'freq': freq})
             except Exception:
-                sys.stderr.write(
+                LOGGER.exception(
                     'Error when inserting into pinyin table. '
-                    + 'count=%(c)s pinyin=%(p)s zi=%(z)s freq=%(f)s\n'
-                    % {'c': count, 'p': pinyin, 'z': zi, 'f': freq})
-                import traceback
-                traceback.print_exc()
+                    'count=%s pinyin=%s zi=%s freq=%s',
+                    count, pinyin, zi, freq)
+        self.db.commit()
+
+    def add_suggestion(self, suggestions, database='main'):
+        '''Add suggestion phrase to database, suggestions is a iterable object
+        Like: [(phrase, freq), (phrase, freq), ...]
+        '''
+        sqlstr = '''
+        INSERT INTO %s.suggestion (phrase, freq) VALUES (:phrase, :freq);
+        ''' % database
+        count = 0
+        for phrase, freq in suggestions:
+            count += 1
+            try:
+                self.db.execute(
+                    sqlstr, {'phrase': phrase, 'freq': freq})
+            except Exception:
+                LOGGER.exception(
+                    'Error when inserting into suggestion table. '
+                    'count=%s phrase=%s freq=%s',
+                    count, phrase, freq)
         self.db.commit()
 
     def optimize_database(self):
@@ -910,6 +925,10 @@ class TabSqliteDb:
             DELETE FROM main.pinyin;
             INSERT INTO main.pinyin SELECT * FROM tmp ORDER BY pinyin ASC, freq DESC;
             DROP TABLE tmp;
+            CREATE TABLE tmp as SELECT * FROM main.suggestion;
+            DELETE FROM main.suggestion;
+            INSERT INTO main.suggestion SELECT * FROM tmp ORDER by phrase ASC, freq DESC;
+            DROP TABLE tmp;
             '''
         self.db.executescript(sqlstr)
         self.db.executescript("VACUUM;")
@@ -922,8 +941,7 @@ class TabSqliteDb:
         function does nothing.
         '''
         if DEBUG_LEVEL > 1:
-            sys.stderr.write("drop_indexes()\n")
-        return
+            LOGGER.debug('drop_indexes()')
 
     def create_indexes(self, _database, _commit=True):
         '''Create indexes for the database.
@@ -938,8 +956,7 @@ class TabSqliteDb:
         be created here (and dropped in “drop_indexes()”).
         '''
         if DEBUG_LEVEL > 1:
-            sys.stderr.write("create_indexes()\n")
-        return
+            LOGGER.debug('create_indexes()')
 
     def big5_code(self, phrase):
         '''
@@ -1066,8 +1083,7 @@ class TabSqliteDb:
             tabkeys_for_like += '%%'
         sqlargs = {'tabkeys': tabkeys_for_like, 'escapechar': escapechar}
         if DEBUG_LEVEL > 1:
-            sys.stderr.write('select_words() sqlstr=%s sqlargs=%s\n'
-                             %(sqlstr, repr(sqlargs)))
+            LOGGER.debug('sqlstr=%s sqlargs=%s', sqlstr, repr(sqlargs))
         unfiltered_results = self.db.execute(sqlstr, sqlargs).fetchall()
         bitmask = None
         if chinese_mode == 0:
@@ -1106,7 +1122,7 @@ class TabSqliteDb:
             candidates=phrase_frequencies.values(),
             chinese_mode=chinese_mode)
         if DEBUG_LEVEL > 1:
-            sys.stderr.write("select_words() best=%s\n" %repr(best))
+            LOGGER.debug('best=%s', repr(best))
         self._phrases_cache[tabkeys] = best
         return best
 
@@ -1121,8 +1137,8 @@ class TabSqliteDb:
             return []
         sqlstr = '''
         SELECT pinyin, zi, freq FROM main.pinyin WHERE pinyin LIKE :tabkeys
-        ORDER BY freq DESC, pinyin ASC
-        ;'''
+        ORDER BY freq DESC, pinyin ASC;
+        '''
         tabkeys_for_like = tabkeys
         if single_wildcard_char:
             tabkeys_for_like = tabkeys_for_like.replace(
@@ -1152,6 +1168,53 @@ class TabSqliteDb:
             candidates=phrase_frequencies,
             chinese_mode=chinese_mode)
 
+    def select_suggestion_candidate(self, prefix=u''):
+        '''
+        Get Chinese phrase matching the prefix from the database.
+        '''
+        if not prefix:
+            return []
+        sqlstr = '''
+        SELECT phrase, freq FROM main.suggestion WHERE phrase LIKE :prefix
+        ORDER BY length(phrase) DESC, freq DESC, phrase ASC;
+        '''
+        prefix_for_like = prefix + '%%'
+        sqlargs = {'prefix': prefix_for_like}
+        results = self.db.execute(sqlstr, sqlargs).fetchall()
+        phrase_frequencies = {}
+        # merge the same phrase in suggestion candidates
+        for phrase, freq in results:
+            if phrase not in phrase_frequencies:
+                phrase_frequencies[phrase] = (phrase, freq)
+            else:
+                phrase_frequencies.update(
+                    [(phrase,
+                      (phrase, max(freq, phrase_frequencies[phrase][1])))])
+        candidates = phrase_frequencies.values()
+        if DEBUG_LEVEL > 1:
+            LOGGER.debug('candidates=%s', repr(candidates))
+        maximum_number_of_candidates = 100
+        engine_name = os.path.basename(self.filename).replace('.db', '')
+
+        if engine_name in [
+                'cangjie3', 'cangjie5', 'cangjie-big',
+                'quick-classic', 'quick3', 'quick5']:
+            code_point_function = self.big5_code
+        else:
+            code_point_function = lambda x: (1)
+
+        return sorted(candidates,
+                      key=lambda x: (
+                          - int(len(x[0])), # longest matches first!
+                          -1*x[1],   # freq descending
+                          code_point_function(x[0][0]),
+                          code_point_function(x[0][1]),
+                          # Unicode codepoint of first character of phrase:
+                          ord(x[0][0]),
+                          # Unicode codepoint of second character of phrase:
+                          ord(x[0][1])
+                      ))[:maximum_number_of_candidates]
+
     def generate_userdb_desc(self):
         '''
         Add a description table to the user database
@@ -1171,8 +1234,7 @@ class TabSqliteDb:
             self.db.execute(sqlstring, ("create-time", ))
             self.db.commit()
         except:
-            import traceback
-            traceback.print_exc()
+            LOGGER.exception('Unexpected error in generate_userdb_desc().')
 
     def init_user_db(self, db_file):
         '''
@@ -1251,8 +1313,7 @@ class TabSqliteDb:
             if res:
                 tp = res.group(1).split(',')
                 return len(tp)
-            else:
-                return 0
+            return 0
         except:
             return 0
 
@@ -1267,7 +1328,7 @@ class TabSqliteDb:
         else:
             goucima = u''
         if DEBUG_LEVEL > 1:
-            sys.stderr.write("get_goucima() goucima=%s\n" %goucima)
+            LOGGER.debug('goucima=%s', goucima)
         return goucima
 
     def parse_phrase(self, phrase):
@@ -1310,9 +1371,7 @@ class TabSqliteDb:
 
         '''
         if DEBUG_LEVEL > 1:
-            sys.stderr.write(
-                'parse_phrase() phrase=%(p)s rules%(r)s\n'
-                % {'p': phrase, 'r': self.rules})
+            LOGGER.debug('phrase=%s rules%s', phrase, self.rules)
         if type(phrase) != type(u''):
             phrase = phrase.decode('UTF-8')
         # Shouldn’t this function try first whether the system database
@@ -1331,15 +1390,14 @@ class TabSqliteDb:
         elif len(phrase) > self.rules['above']:
             rule = self.rules[self.rules['above']]
         else:
-            sys.stderr.write(
-                'No rule for this phrase length. phrase=%(p)s rules=%(r)s\n'
-                %{'p': phrase, 'r': self.rules})
+            LOGGER.debug(
+                'No rule for this phrase length. phrase=%s rules=%s',
+                phrase, self.rules)
             return u''
         if len(rule) > self._mlen:
-            sys.stderr.write(
+            LOGGER.debug(
                 'Rule exceeds maximum key length. '
-                + 'rule=%(r)s self._mlen=%(m)s\n'
-                %{'r': rule, 'm': self._mlen})
+                'rule=%s self._mlen=%s', rule, self._mlen)
             return u''
         tabkeys = u''
         for (zi, ma) in rule:
@@ -1352,7 +1410,7 @@ class TabSqliteDb:
                 return u''
             tabkeys += tabkey
         if DEBUG_LEVEL > 1:
-            sys.stderr.write("parse_phrase() tabkeys=%s\n" %tabkeys)
+            LOGGER.debug('tabkeys=%s', tabkeys)
         return tabkeys
 
     def is_in_system_database(self, tabkeys=u'', phrase=u''):
@@ -1361,9 +1419,7 @@ class TabSqliteDb:
         with a key sequence *starting* with “tabkeys”.
         '''
         if DEBUG_LEVEL > 1:
-            sys.stderr.write(
-                'is_in_system_database() tabkeys=%(t)s phrase=%(p)s\n'
-                % {'t': tabkeys, 'p': phrase})
+            LOGGER.debug('tabkeys=%s phrase=%s', tabkeys, phrase)
         if not tabkeys or not phrase:
             return False
         sqlstr = '''
@@ -1373,11 +1429,9 @@ class TabSqliteDb:
         sqlargs = {'tabkeys': tabkeys+'%%', 'phrase': phrase}
         results = self.db.execute(sqlstr, sqlargs).fetchall()
         if DEBUG_LEVEL > 1:
-            sys.stderr.write(
-                'is_in_system_database() tabkeys=%(t)s phrase=%(p)s '
-                % {'t': tabkeys, 'p': phrase}
-                + 'results=%(r)s\n'
-                % {'r': results})
+            LOGGER.debug(
+                'tabkeys=%s phrase=%s results=%s',
+                tabkeys, phrase, results)
         return bool(results)
 
     def user_frequency(self, tabkeys=u'', phrase=u''):
@@ -1393,9 +1447,7 @@ class TabSqliteDb:
 
         '''
         if DEBUG_LEVEL > 1:
-            sys.stderr.write(
-                'user_frequency() tabkeys=%(t)s phrase=%(p)s\n'
-                % {'t': tabkeys, 'p': phrase})
+            LOGGER.debug('tabkeys=%s phrase=%s', tabkeys, phrase)
         if not tabkeys or not phrase:
             return 0
         sqlstr = '''
@@ -1405,7 +1457,7 @@ class TabSqliteDb:
         sqlargs = {'tabkeys': tabkeys, 'phrase': phrase}
         result = self.db.execute(sqlstr, sqlargs).fetchall()
         if DEBUG_LEVEL > 1:
-            sys.stderr.write("user_frequency() result=%s\n" %result)
+            LOGGER.debug('result=%s', result)
         if result:
             return result[0][0]
         return 0
@@ -1419,9 +1471,7 @@ class TabSqliteDb:
         there.
         '''
         if DEBUG_LEVEL > 1:
-            sys.stderr.write(
-                'check_phrase_internal() tabkey=%(t)s phrase=%(p)s\n'
-                % {'t': tabkeys, 'p': phrase})
+            LOGGER.debug('tabkey=%s phrase=%s', tabkeys, phrase)
         if type(phrase) != type(u''):
             phrase = phrase.decode('utf8')
         if type(tabkeys) != type(u''):
@@ -1523,9 +1573,9 @@ class TabSqliteDb:
     def extract_user_phrases(
             self, database_file='', old_database_version='0.0'):
         '''extract user phrases from database'''
-        sys.stderr.write(
+        LOGGER.debug(
             'Trying to recover the phrases from the old, '
-            + 'incompatible database.\n')
+            'incompatible database.')
         try:
             db = sqlite3.connect(database_file)
             db.execute('PRAGMA wal_checkpoint;')
@@ -1539,51 +1589,48 @@ class TabSqliteDb:
                 db.close()
                 phrases = sorted(
                     phrases, key=lambda x: (x[0], x[1], x[2], x[3]))
-                sys.stderr.write(
-                    'Recovered phrases from the old database: phrases=%s\n'
-                    % repr(phrases))
-                return phrases[:]
-            else:
-                # database is very old, it may still use many columns
-                # of type INTEGER for the tabkeys. Therefore, ignore
-                # the tabkeys in the database and try to get them
-                # from the system database instead.
-                phrases = []
-                results = db.execute(
-                    'SELECT phrase, sum(user_freq) '
-                    + 'FROM phrases GROUP BY phrase;'
-                ).fetchall()
-                for result in results:
-                    sqlstr = '''
-                    SELECT tabkeys FROM main.phrases WHERE phrase = :phrase
-                    ORDER BY length(tabkeys) DESC;
-                    '''
-                    sqlargs = {'phrase': result[0]}
-                    tabkeys_results = self.db.execute(
-                        sqlstr, sqlargs).fetchall()
-                    if tabkeys_results:
-                        phrases.append(
-                            (tabkeys_results[0][0], result[0], 0, result[1]))
-                    else:
-                        # No tabkeys for that phrase could not be
-                        # found in the system database.  Try to get
-                        # tabkeys by calling self.parse_phrase(), that
-                        # might return something if the table has
-                        # rules to construct user defined phrases:
-                        tabkeys = self.parse_phrase(result[0])
-                        if tabkeys:
-                            # for user defined phrases, the “freq”
-                            # column is -1:
-                            phrases.append((tabkeys, result[0], -1, result[1]))
-                db.close()
-                phrases = sorted(
-                    phrases, key=lambda x: (x[0], x[1], x[2], x[3]))
-                sys.stderr.write(
-                    'Recovered phrases from the very old database: '
-                    + 'phrases=%s\n'
-                    % repr(phrases))
+                LOGGER.debug(
+                    'Recovered phrases from the old database: phrases=%s',
+                    repr(phrases))
                 return phrases[:]
+            # database is very old, it may still use many columns
+            # of type INTEGER for the tabkeys. Therefore, ignore
+            # the tabkeys in the database and try to get them
+            # from the system database instead.
+            phrases = []
+            results = db.execute(
+                'SELECT phrase, sum(user_freq) '
+                + 'FROM phrases GROUP BY phrase;'
+            ).fetchall()
+            for result in results:
+                sqlstr = '''
+                SELECT tabkeys FROM main.phrases WHERE phrase = :phrase
+                ORDER BY length(tabkeys) DESC;
+                '''
+                sqlargs = {'phrase': result[0]}
+                tabkeys_results = self.db.execute(
+                    sqlstr, sqlargs).fetchall()
+                if tabkeys_results:
+                    phrases.append(
+                        (tabkeys_results[0][0], result[0], 0, result[1]))
+                else:
+                    # No tabkeys for that phrase could not be
+                    # found in the system database.  Try to get
+                    # tabkeys by calling self.parse_phrase(), that
+                    # might return something if the table has
+                    # rules to construct user defined phrases:
+                    tabkeys = self.parse_phrase(result[0])
+                    if tabkeys:
+                        # for user defined phrases, the “freq”
+                        # column is -1:
+                        phrases.append((tabkeys, result[0], -1, result[1]))
+            db.close()
+            phrases = sorted(
+                phrases, key=lambda x: (x[0], x[1], x[2], x[3]))
+            LOGGER.debug(
+                'Recovered phrases from the very old database: '
+                'phrases=%s', repr(phrases))
+            return phrases[:]
         except:
-            import traceback
-            traceback.print_exc()
+            LOGGER.exception('Unexpected error in extract_user_phrases()')
             return []
